from __future__ import division, print_function
import numpy as np

"""
Generates a multi-harmonic dictionary based on a time vector (t in Nx1) and a fundamental frequency (f0)
"""
def harmonic_dictionary(t, f0, M=1):
    N = len(t)
    Phi = np.ones(shape=(N, 2*M+1))
    for k in range(0, M):
        Phi[:, k+1] = np.cos(2.0*np.pi*t*f0*(k+1))
        Phi[:, M+k+1] = np.sin(2.0*np.pi*t*f0*(k+1))
    return Phi



from __future__ import division, print_function
import numpy as np


def find_beta_WMCC(y, Phi, dy, max_inner_iterations=1, max_outer_iterations=100, stopping_tol=1.01, debug=False, full_output=False):
    """
    Find coefficient vector (beta in R^{Mx1}) that better fits data 
    vector (y in R^{Nx1}) to dictionary matrix (Phi in R^{NxM}) 
        y approx np.dot(Phi, beta)
    Objective function can be the Weighted Maximum Correntropy Criterion 
    (WMCC), Weighted Least Squares (WLS) or Ordinary Least Squares (OLS)
    
    max_inner_iterations -- Number of iterations for fixed point and 
    kernel size adaption routines
    max_outer_iterations -- Number of iterations for the whole routine
    stopping_tol -- Stopping criteria check for convergence
    
    Returns beta, cost function evolution vector, and kernel size 
    evolution vector. The latter two are only relevant for the WMCC
    
    TODO: Study kernel size normalizations for WMCC
    """

    N, M = Phi.shape
    beta = np.zeros(shape=(M,))
    dy2 = np.power(dy, 2.0)
    # Initial kernel size:
    iqr_y = np.percentile(y, 75) - np.percentile(y, 25)
    ks_silverman = 1.0592*np.minimum(np.std(y), iqr_y/1.34)*np.power(N, -0.2)
    # Optimize log(ks) instead of ks (positive constraint)
    log_ks = np.log(10.0*ks_silverman)  # Very conservative initial value
    cost_history = np.zeros(shape=(max_outer_iterations,))
    kernel_size_history = np.zeros(shape=(max_outer_iterations,))
    # ADADELTA parameters
    rho = 0.9
    eps = 1e-6 
    ks_grad2_hist = 0.0
    ks_dx2_hist = 0.01 # default should be zero but >0 improves convergence speed
    # Main routine
    kernel_size_history[0] = np.exp(log_ks)
    ks2 = dy2 + np.exp(2*log_ks)
    error = y - np.dot(Phi, beta)
    PhiT = np.transpose(Phi)
    for i in range(0, max_outer_iterations):
        # Fixed point routine to update beta
        ks = np.sqrt(ks2)
        for j in range(0, max_inner_iterations):
            cost = np.sum(np.divide(np.exp(-0.5*np.divide(np.power(error, 2.0), ks2)), ks))
            cost_history[i] = cost/(N*np.sqrt(2.0*np.pi))
            #cost_history[i] = 1.0 - cost/np.sum(1.0/ks)
            if i > 0 and cost_history[i] <= cost_history[i-1]*stopping_tol:
                break
            W = np.divide(cost, ks2)  # diagonal matrix
            beta = np.linalg.solve(np.dot(PhiT*W, Phi), np.dot(PhiT*W, y))
            """
            TODO: Inversion could be not feasible, check conditioning, add regularization...
            """
            error = y - np.dot(Phi, beta)
            #grad = np.dot(A.T, np.dot(W, e))/(N*np.sqrt(2.0*np.pi))
            #beta += lr*grad
        if i > 0 and cost_history[i] <= cost_history[i-1]*stopping_tol:  # More sophisticated stopping criteria could be tried
            if debug:
                print("Breaking at iteration %d" % i)
            break    
        # Gradient descent with ADADELTA to update the kernel size
        e2 = np.power(error, 2.0)
        for j in range(0, max_inner_iterations):
            ks = np.sqrt(ks2)
            cost = np.sum(np.divide(np.exp(-0.5*np.divide(e2, ks2)), ks))
            cost_history[i] = cost/(N*np.sqrt(2.0*np.pi))
            #den = np.sum(1.0/ks)
            #cost_history[i] = 1.0 - cost/den 
            if i > 0 and cost_history[i] <= cost_history[i-1]*stopping_tol:
                break
            W = np.divide(cost, ks2)
            grad = np.sum(np.multiply(W, np.divide(e2, ks2) - 1.0))/(N*np.sqrt(2.0*np.pi))
            #ks3 = np.sum(1.0/(ks2*ks))
            #grad = np.sum(np.multiply(W, np.divide(e2, ks2) - 1.0))/den + cost*ks3/den**2
            grad = grad*np.exp(2.0*log_ks)
            ks_grad2_hist = rho*ks_grad2_hist + (1.0-rho)*grad**2
            RMSgt2 = np.sqrt(ks_grad2_hist + eps)
            RMSdx = np.sqrt(ks_dx2_hist + eps)
            dlog_ks = (RMSdx/RMSgt2)*grad
            ks_dx2_hist = rho*ks_dx2_hist + (1.0-rho)*dlog_ks**2
            log_ks += dlog_ks
            kernel_size_history[i] = np.exp(log_ks)
            ks2 = dy2 + np.exp(2*log_ks)
    if full_output:
        return beta, cost_history[:i], kernel_size_history[:i]
    else:
        return beta, cost_history[:i][-1]
    

def find_beta_OLS(y, Phi, dy=None):
    """
    Ordinary least squares (OLS) regression for benchmark purposes
    """
    R = np.dot(Phi.T, Phi)
    P = np.dot(Phi.T, y)
    beta = np.linalg.solve(R, P)
    return beta, np.dot(P.T, beta)/(np.var(y)*len(y))


def find_beta_WLS(y, Phi, dy):
    """
    Weighted least squares (WLS) regression for benchmark purposes
    """
    W = np.power(dy, -2.0)
    #W_mat = np.diag(dy2)
    R = np.dot(Phi.T*W, Phi)
    P = np.dot(Phi.T*W, y)
    beta = np.linalg.solve(R, P)
    return beta, np.dot(P.T, beta)/np.sum(np.multiply(np.power(y, 2.0), W))
    




from __future__ import division, print_function
import numpy as np
#from scipy.stats import genextreme as gev
from scipy.stats import gumbel_r
from .regression import find_beta_WMCC, find_beta_OLS, find_beta_WLS
from .dictionary import harmonic_dictionary
#from multiprocessing import Pool
#from multiprocess import Pool
#from functools import partial


#def unwrap_self(arg, **kwarg):
#    return periodogram.compute_per_ordinate(*arg, **kwarg)

class periodogram:
    def __init__(self, method='WMCC', M=1, n_jobs=1):
        """
        Class for multiharmonic periodogram computation
        M -- number of harmonic components used to fit the data
        method -- method used to perform the fit, options are OLS, WLS and WMCC
        n_jobs -- number of parallel jobs for periodogram computation, greater than 1
        """
        if n_jobs < 1:
            raise ValueError("Number of jobs must be greater than 0")
        self.method = method
        self.M = M
        self.local_max_index = None
        self.freq = None
        self.per = None
        self.n_jobs = n_jobs
        if self.method == 'WMCC':
            self.get_cost = find_beta_WMCC
        elif self.method == 'WLS':
            self.get_cost = find_beta_WLS
        elif self.method == 'OLS':
            self.get_cost = find_beta_OLS
        
    def fit(self, t, y, dy):
        self.t = t
        self.y = y #- np.mean(y)
        self.dy = dy
        self.T = t[-1] - t[0]
    
    def get_best_frequency(self):
        return self.freq[self.best_local_max[0]], self.per[self.best_local_max[0]]
        
    def get_best_frequencies(self):
        """
        Returns the best n_local_max frequencies and their periodogram 
        values, sorted by per
        """
        return self.freq[self.best_local_max], self.per[self.best_local_max]
        
    def get_periodogram(self):
        return self.freq, self.per
        
    def compute_per_ordinate(self, f):
        Phi = harmonic_dictionary(self.t, f, self.M)
        _, per = self.get_cost(self.y, Phi, self.dy)
        return per
    
    def grid_search(self, fmin=0.0, fmax=1.0, fres_coarse=1.0, fres_fine=0.1, n_local_max=10):
        """ 
        Computes self.method over a grid of frequencies specified by
        fmin -- starting frequency
        fmax -- stopping frequency
        fres_coarse -- step size in the frequency grid, note that the 
        actual frequency step is fres_coarse/self.T, where T is the 
        total time span of the time series
        
        Then it refines (fine-tune) the estimation for a given number of local maxima:
        n_local_max -- number of local maxima to refine
        fres_fine -- oversampling factor for the fine-tuning step
        """
        self.fres_coarse = fres_coarse
        freq = np.arange(np.amax([fmin, fres_coarse/self.T]), fmax, step=fres_coarse/self.T)
        Nf = len(freq)
        per = np.zeros(shape=(Nf,))
        
        """
        #partial_job = partial(self.compute_per_ordinate)
        if self.n_jobs <= 1:
            m = map
        else:
            pool = Pool(self.n_jobs)
            m = pool.map
        per = list(m(self.compute_per_ordinate, freq))
        if self.n_jobs > 1:
            pool.close()
            pool.join()
        per = np.asarray(per, dtype=float)
        """        
        for k in range(0, Nf):
            #per[k] = self.compute_per_ordinate(freq[k], self.t, self.y, self.dy, self.M)
            Phi = harmonic_dictionary(self.t, freq[k], self.M)
            _, per[k] = self.get_cost(self.y, Phi, self.dy)
        # Find the local minima and do analysis with finer frequency step
        local_max_index = []
        for k in range(1, Nf-1):
            if per[k-1] < per[k] and per[k+1] < per[k]:
                local_max_index.append(k)
        local_max_index = np.array(local_max_index)
        best_local_max = local_max_index[np.argsort(per[local_max_index])][::-1][:n_local_max]
        #print(freq[best_local_max])
        # Do finetuning
        for j in range(0, n_local_max):
            freq_fine = freq[best_local_max[j]] - fres_coarse/self.T
            for k in range(0, int(2.0*fres_coarse/fres_fine)):
                Phi = harmonic_dictionary(self.t, freq_fine, self.M)
                _, cost = self.get_cost(self.y, Phi, self.dy)
                if cost > per[best_local_max[j]]:
                    per[best_local_max[j]] = cost
                    freq[best_local_max[j]] = freq_fine
                freq_fine += fres_fine/self.T
        # Sort them
        idx = np.argsort(per[best_local_max])[::-1]
        self.best_local_max= best_local_max[idx]
        self.freq = freq
        self.per = per
        return freq, per

    def get_confidence(self, per_value):
        """
        Computes the confidence for a given periodogram value
        """
        return gumbel_r.cdf(per_value, loc=self.param[0], scale=self.param[1])
    
    def get_FAP(self, p):
        """
        Computes the periodogram value associated to FAP=p
        """
        return self.param[0] - self.param[1]*np.log(-np.log(1.0-p))

    def fit_extreme_cdf(self, n_bootstrap=10, n_frequencies=10):
        """
        Perform false alarm probability (FAP) computation based on 
        generalized extreme value (gev) statistics. 
        
        n_bootstrap -- the number of bootstrap repetitions of time 
        series (t,y,dy)
        n_frequencies -- the number of frequencies to search for maxima, 
        it is a subset of self.freq
        
        Returns the maxima for each bootstrap repetition and the
        parameters resulting from the fit
        
        Reference:
        Süveges, M. "False Alarm Probability based on bootstrap and 
        extreme-value methods for periodogram peaks." ADA7-Seventh 
        Conference on Astronomical Data Analysis. Vol. 1. 2012.
        """
        
        y = self.y.copy()
        dy = self.dy.copy()
        #K = int(1.0/self.fres_coarse)  # oversampling factor 
        K = 1
        N = len(self.t)
        Nf = len(self.freq)
        # Sensible limits for the number of frequencies
        if n_frequencies > Nf: 
            n_frequencies = Nf
        if n_frequencies < 2*Nf/(K*N):
            n_frequencies = int(2*Nf/(K*N))
        idx = np.random.randint(0, N, (n_bootstrap, N))
        maxima_realization = np.zeros(shape=(n_bootstrap,))
        # Find the maxima
        for i in range(0, n_bootstrap):  # bootstrap
            random_freq = np.random.permutation(Nf)[:n_frequencies]
            per_gev = np.zeros(shape=(n_frequencies,))
            for k in range(0, n_frequencies):
                Phi = harmonic_dictionary(self.t, self.freq[random_freq[k]], self.M)
                _, per_gev[k] = self.get_cost(y[idx[i]], Phi, dy[idx[i]])
            maxima_realization[i] = np.amax(per_gev)
        # Fit the GEV parameters
        self.param = gumbel_r.fit(maxima_realization)
        #return self.param[0] - self.param[1]*np.log(-np.log(1.0-p))
        return maxima_realization, self.param



from .regression import find_beta_WMCC, find_beta_OLS, find_beta_WLS
from .dictionary import harmonic_dictionary
from .generator import contaminate_time_series, irregular_sampling, trigonometric_model
from .periodogram import periodogram

__version__ = '0.5'



from __future__ import division, print_function
import numpy as np


def irregular_sampling(T, N, rseed=None):
    """
    Generates an irregularly sampled time vector with specified number of 
    samples (N) and time span (T)
    """
    N = int(N)
    np.random.seed(rseed)
    sampling_period = (T/N)
    t = np.linspace(0, T, num=5*N)
    # First we add jitter
    t[1:-1] += sampling_period*0.5*np.random.randn(5*N-2)
    # Then we do a random permutation and keep only N points 
    P = np.random.permutation(5*N)
    return np.sort(t[P[:N]])
    

def trigonometric_model(t, f0, A):
    """
    Generates a normalized trigonometric model, A corresponds to a vector
    having the amplitudes of the model and t is a time vector, f0 is the 
    fundamental frequency of the model
    """
    y = 0.0
    M = len(A)
    var_y = 0.0
    for k in range(0, M):
        y += A[k]*np.sin(2.0*np.pi*t*f0*(k+1))
        var_y += 0.5*A[k]**2
    return y/np.sqrt(var_y)


def contaminate_time_series(t, y, SNR=10.0, red_noise_ratio=0.5, outlier_ratio=0.0, rseed=None):
    """
    A function that takes an irregulary sampled time series (t, y) 
    and contaminates it, simulating the noise present in light curves. 
    It returns (y, y_noisy, s), where s are the uncertainties 
    associated to the magnitudes (y_noisy). y will be reescaled to fit a 
    given SNR. 
    
    SNR -- Signal to Noise ratio of the resulting time series in dB. SNR }
    is defined as SNR = 10*log(var_signal/var_noise), so
     SNR var_signal/var_noise
     10  10
      7   5
      3   2
      0   1
     -3   0.5
     -7   0.2
    -10   0.1
    outlier_ratio -- percentage of outlier data points, should be in [0, 1]
    red_noise_ratio -- How large is the variance of the red noise with
    respect to the white noise component. The red noise is not accounted 
    into the uncertainties (s). Setting red_noise_ratio=0.0, would yield
    "perfect" errorbars.
    """
    if outlier_ratio < 0.0 or outlier_ratio > 1.0:
        raise ValueError("Outlier ratio should be in [0 , 1]")
    np.random.seed(rseed)
    N = len(t)
    # First we generate s from a Gamma distribution
    b = 1.0
    k = 3.0
    s = np.random.gamma(shape=k, scale=b, size=(N,))
    mean_ds_squared = k*(1+k)*b**2   # From E[s**2]
    # Draw a heteroscedastic white noise vector
    white_noise = np.random.multivariate_normal(np.zeros(N,), np.diag(s**2))
    # Now we generate a colored noise vector which is unaccounted in s
    phi = 0.5  # correlation coefficient
    red_noise_variance = mean_ds_squared*red_noise_ratio
    red_noise = np.random.randn(N)*np.sqrt(red_noise_variance)
    for i in range(1, N):
        red_noise[i] = phi*red_noise[i-1] + np.sqrt(1 - phi**2)*red_noise[i]
    # The final noise vector
    noise = white_noise + red_noise 
    var_noise = mean_ds_squared + red_noise_variance
    SNR_unitless = 10.0**(SNR/10.0)
    y = np.sqrt(SNR_unitless*var_noise)*y
    y_noisy = y + noise
    # Add outliers with a certain percentage
    rperm = np.where(np.random.uniform(size=N) < outlier_ratio)[0]    
    outlier = np.random.uniform(5.0*np.std(y), 10.0*np.std(y), size=len(rperm))
    y_noisy[rperm] += outlier
    return y, y_noisy, s



import re
import io
import os
from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()

def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def version(path):
    """
    https://packaging.python.org/en/latest/single_source_version.html
    """
    version_file = read(path)
    version_match = re.search(r"""^__version__ = ['"]([^'"]*)['"]""",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")
    
setup(
    name = 'P4J',
    packages = ['P4J'], 
    version = version('P4J/__init__.py'),
    description = 'Periodic light curve analysis tools based on Information Theory',
    long_description=readme(),
    author = 'Pablo Huijse',
    author_email = 'pablo.huijse@gmail.com',
    license='MIT',
    url = 'https://github.com/phuijse/P4J', 
    download_url = 'https://github.com/phuijse/P4J/tarball/stable', 
    keywords = ['astronomy periodic time series correntropy'], 
    install_requires=[
        'numpy',
        'scipy'
    ],
    classifiers = [
        'Natural Language :: English',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Astronomy',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)
